/*
 * Copyright 2019-2020 Ernst Jan Plugge <rmc@dds.nl>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.smouldering_durtles.wk.views;

import android.content.Context;
import android.util.AttributeSet;
import android.widget.TableLayout;

import androidx.lifecycle.LifecycleOwner;

import com.smouldering_durtles.wk.GlobalSettings;
import com.smouldering_durtles.wk.R;
import com.smouldering_durtles.wk.db.model.Subject;
import com.smouldering_durtles.wk.enums.ActiveTheme;
import com.smouldering_durtles.wk.livedata.LiveFirstTimeSetup;
import com.smouldering_durtles.wk.livedata.LiveLevelDuration;
import com.smouldering_durtles.wk.livedata.LiveTimeLine;
import com.smouldering_durtles.wk.model.TimeLine;
import com.smouldering_durtles.wk.proxy.ViewProxy;
import com.smouldering_durtles.wk.util.ThemeUtil;

import javax.annotation.Nullable;

import static com.smouldering_durtles.wk.util.ObjectSupport.safe;

/**
 * A custom view that shows a breakdown of available lessons and reviews.
 */
public final class LessonReviewBreakdownView extends TableLayout {
    private final ViewProxy headerRad = new ViewProxy();
    private final ViewProxy headerKan = new ViewProxy();
    private final ViewProxy headerVoc = new ViewProxy();
    private final ViewProxy lessonCurrent = new ViewProxy();
    private final ViewProxy lessonCurrentRad = new ViewProxy();
    private final ViewProxy lessonCurrentKan = new ViewProxy();
    private final ViewProxy lessonCurrentVoc = new ViewProxy();
    private final ViewProxy lessonPast = new ViewProxy();
    private final ViewProxy lessonPastRad = new ViewProxy();
    private final ViewProxy lessonPastKan = new ViewProxy();
    private final ViewProxy lessonPastVoc = new ViewProxy();
    private final ViewProxy reviewCurrent = new ViewProxy();
    private final ViewProxy reviewCurrentRad = new ViewProxy();
    private final ViewProxy reviewCurrentKan = new ViewProxy();
    private final ViewProxy reviewCurrentVoc = new ViewProxy();
    private final ViewProxy reviewPast = new ViewProxy();
    private final ViewProxy reviewPastRad = new ViewProxy();
    private final ViewProxy reviewPastKan = new ViewProxy();
    private final ViewProxy reviewPastVoc = new ViewProxy();

    /**
     * The constructor.
     *
     * @param context Android context
     */
    public LessonReviewBreakdownView(final Context context) {
        super(context);
        safe(this::init);
    }

    /**
     * The constructor.
     *
     * @param context Android context
     * @param attrs attribute set
     */
    public LessonReviewBreakdownView(final Context context, final AttributeSet attrs) {
        super(context, attrs);
        safe(this::init);
    }

    /**
     * Initialize the view by observing the relevant LiveData instances.
     */
    private void init() {
        inflate(getContext(), R.layout.lesson_review_breakdown, this);
        setColumnStretchable(0, true);
        setBackgroundResource(R.drawable.main_activity_view_background);

        headerRad.setDelegate(this, R.id.headerRad);
        headerKan.setDelegate(this, R.id.headerKan);
        headerVoc.setDelegate(this, R.id.headerVoc);
        lessonCurrent.setDelegate(this, R.id.lessonCurrent);
        lessonCurrentRad.setDelegate(this, R.id.lessonCurrentRad);
        lessonCurrentKan.setDelegate(this, R.id.lessonCurrentKan);
        lessonCurrentVoc.setDelegate(this, R.id.lessonCurrentVoc);
        lessonPast.setDelegate(this, R.id.lessonPast);
        lessonPastRad.setDelegate(this, R.id.lessonPastRad);
        lessonPastKan.setDelegate(this, R.id.lessonPastKan);
        lessonPastVoc.setDelegate(this, R.id.lessonPastVoc);
        reviewCurrent.setDelegate(this, R.id.reviewCurrent);
        reviewCurrentRad.setDelegate(this, R.id.reviewCurrentRad);
        reviewCurrentKan.setDelegate(this, R.id.reviewCurrentKan);
        reviewCurrentVoc.setDelegate(this, R.id.reviewCurrentVoc);
        reviewPast.setDelegate(this, R.id.reviewPast);
        reviewPastRad.setDelegate(this, R.id.reviewPastRad);
        reviewPastKan.setDelegate(this, R.id.reviewPastKan);
        reviewPastVoc.setDelegate(this, R.id.reviewPastVoc);

        headerRad.setTextColor(ActiveTheme.getSubjectTypeTextColors()[0]);
        headerRad.setBackgroundColor(ActiveTheme.getSubjectTypeBackgroundColors()[0]);
        headerKan.setTextColor(ActiveTheme.getSubjectTypeTextColors()[1]);
        headerKan.setBackgroundColor(ActiveTheme.getSubjectTypeBackgroundColors()[1]);
        headerVoc.setTextColor(ActiveTheme.getSubjectTypeTextColors()[2]);
        headerVoc.setBackgroundColor(ActiveTheme.getSubjectTypeBackgroundColors()[2]);

        lessonCurrentRad.setTextColor(ActiveTheme.getSubjectTypeTextColors()[0]);
        lessonCurrentRad.setBackgroundColor(ActiveTheme.getSubjectTypeBackgroundColors()[0]);
        lessonCurrentKan.setTextColor(ActiveTheme.getSubjectTypeTextColors()[1]);
        lessonCurrentKan.setBackgroundColor(ActiveTheme.getSubjectTypeBackgroundColors()[1]);
        lessonCurrentVoc.setTextColor(ActiveTheme.getSubjectTypeTextColors()[2]);
        lessonCurrentVoc.setBackgroundColor(ActiveTheme.getSubjectTypeBackgroundColors()[2]);

        lessonPastRad.setTextColor(ActiveTheme.getSubjectTypeTextColors()[0]);
        lessonPastRad.setBackgroundColor(ActiveTheme.getSubjectTypeBackgroundColors()[0]);
        lessonPastKan.setTextColor(ActiveTheme.getSubjectTypeTextColors()[1]);
        lessonPastKan.setBackgroundColor(ActiveTheme.getSubjectTypeBackgroundColors()[1]);
        lessonPastVoc.setTextColor(ActiveTheme.getSubjectTypeTextColors()[2]);
        lessonPastVoc.setBackgroundColor(ActiveTheme.getSubjectTypeBackgroundColors()[2]);

        reviewCurrentRad.setTextColor(ActiveTheme.getSubjectTypeTextColors()[0]);
        reviewCurrentRad.setBackgroundColor(ActiveTheme.getSubjectTypeBackgroundColors()[0]);
        reviewCurrentKan.setTextColor(ActiveTheme.getSubjectTypeTextColors()[1]);
        reviewCurrentKan.setBackgroundColor(ActiveTheme.getSubjectTypeBackgroundColors()[1]);
        reviewCurrentVoc.setTextColor(ActiveTheme.getSubjectTypeTextColors()[2]);
        reviewCurrentVoc.setBackgroundColor(ActiveTheme.getSubjectTypeBackgroundColors()[2]);

        reviewPastRad.setTextColor(ActiveTheme.getSubjectTypeTextColors()[0]);
        reviewPastRad.setBackgroundColor(ActiveTheme.getSubjectTypeBackgroundColors()[0]);
        reviewPastKan.setTextColor(ActiveTheme.getSubjectTypeTextColors()[1]);
        reviewPastKan.setBackgroundColor(ActiveTheme.getSubjectTypeBackgroundColors()[1]);
        reviewPastVoc.setTextColor(ActiveTheme.getSubjectTypeTextColors()[2]);
        reviewPastVoc.setBackgroundColor(ActiveTheme.getSubjectTypeBackgroundColors()[2]);
    }

    /**
     * Set the lifecycle owner for this view, to hook LiveData observers to.
     *
     * @param lifecycleOwner the lifecycle owner
     */
    public void setLifecycleOwner(final LifecycleOwner lifecycleOwner) {
        safe(() -> LiveTimeLine.getInstance().observe(lifecycleOwner, t -> safe(() -> update(t))));
    }

    /**
     * Update the table based on the latest timeline data.
     *
     * @param timeLine the timeline
     */
    private void update(final @Nullable TimeLine timeLine) {
        if (timeLine == null) {
            setVisibility(GONE);
            return;
        }

        final int numLessons = timeLine.getNumAvailableLessons();
        final int numReviews = timeLine.getNumAvailableReviews();

        if (numLessons == 0 && numReviews == 0
                || LiveFirstTimeSetup.getInstance().get() == 0
                || !GlobalSettings.Dashboard.getShowLessonReviewBreakdown()) {
            setVisibility(GONE);
            return;
        }

        setVisibility(VISIBLE);

        final int userLevel = LiveLevelDuration.getInstance().get().getLevel();

        int lessonCurrentRadCount = 0;
        int lessonCurrentKanCount = 0;
        int lessonCurrentVocCount = 0;
        int lessonPastRadCount = 0;
        int lessonPastKanCount = 0;
        int lessonPastVocCount = 0;

        for (final Subject subject: timeLine.getAvailableLessons()) {
            if (subject.getLevel() == userLevel && subject.getType().isRadical()) {
                lessonCurrentRadCount++;
            }
            if (subject.getLevel() == userLevel && subject.getType().isKanji()) {
                lessonCurrentKanCount++;
            }
            if (subject.getLevel() == userLevel && subject.getType().isVocabulary()) {
                lessonCurrentVocCount++;
            }
            if (subject.getLevel() == userLevel && subject.getType().isKanaVocabulary()) {
                lessonCurrentVocCount++;
            }
            if (subject.getLevel() < userLevel && subject.getType().isRadical()) {
                lessonPastRadCount++;
            }
            if (subject.getLevel() < userLevel && subject.getType().isKanji()) {
                lessonPastKanCount++;
            }
            if (subject.getLevel() < userLevel && subject.getType().isVocabulary()) {
                lessonPastVocCount++;
            }
            if (subject.getLevel() < userLevel && subject.getType().isKanaVocabulary()) {
                lessonPastVocCount++;
            }
        }

        lessonCurrent.setVisibility(lessonCurrentRadCount != 0 || lessonCurrentKanCount != 0 || lessonCurrentVocCount != 0);
        lessonCurrentRad.setTextOrBlankIfZero(lessonCurrentRadCount);
        lessonCurrentKan.setTextOrBlankIfZero(lessonCurrentKanCount);
        lessonCurrentVoc.setTextOrBlankIfZero(lessonCurrentVocCount);

        lessonPast.setVisibility(lessonPastRadCount != 0 || lessonPastKanCount != 0 || lessonPastVocCount != 0);
        lessonPastRad.setTextOrBlankIfZero(lessonPastRadCount);
        lessonPastKan.setTextOrBlankIfZero(lessonPastKanCount);
        lessonPastVoc.setTextOrBlankIfZero(lessonPastVocCount);

        int reviewCurrentRadCount = 0;
        int reviewCurrentKanCount = 0;
        int reviewCurrentVocCount = 0;
        int reviewPastRadCount = 0;
        int reviewPastKanCount = 0;
        int reviewPastVocCount = 0;

        for (final Subject subject: timeLine.getAvailableReviews()) {
            if (subject.getLevel() == userLevel && subject.getType().isRadical()) {
                reviewCurrentRadCount++;
            }
            if (subject.getLevel() == userLevel && subject.getType().isKanji()) {
                reviewCurrentKanCount++;
            }
            if (subject.getLevel() == userLevel && subject.getType().isVocabulary()) {
                reviewCurrentVocCount++;
            }
            if (subject.getLevel() == userLevel && subject.getType().isKanaVocabulary()) {
                reviewCurrentVocCount++;
            }
            if (subject.getLevel() < userLevel && subject.getType().isRadical()) {
                reviewPastRadCount++;
            }
            if (subject.getLevel() < userLevel && subject.getType().isKanji()) {
                reviewPastKanCount++;
            }
            if (subject.getLevel() < userLevel && subject.getType().isVocabulary()) {
                reviewPastVocCount++;
            }
            if (subject.getLevel() < userLevel && subject.getType().isKanaVocabulary()) {
                reviewPastVocCount++;
            }
        }

        reviewCurrent.setVisibility(reviewCurrentRadCount != 0 || reviewCurrentKanCount != 0 || reviewCurrentVocCount != 0);
        reviewCurrentRad.setTextOrBlankIfZero(reviewCurrentRadCount);
        reviewCurrentKan.setTextOrBlankIfZero(reviewCurrentKanCount);
        reviewCurrentVoc.setTextOrBlankIfZero(reviewCurrentVocCount);

        reviewPast.setVisibility(reviewPastRadCount != 0 || reviewPastKanCount != 0 || reviewPastVocCount != 0);
        reviewPastRad.setTextOrBlankIfZero(reviewPastRadCount);
        reviewPastKan.setTextOrBlankIfZero(reviewPastKanCount);
        reviewPastVoc.setTextOrBlankIfZero(reviewPastVocCount);
    }
}
